<?php
/**
 * A general file-based stack class
 *
 * @author		Roland Haeder <webmaster@ship-simu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2013 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.ship-simu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
class BaseFileStack extends BaseStacker {
	/**
	 * Magic for this stack
	 */
	const STACK_MAGIC = 'STACKv0.1';

	/**
	 * Protected constructor
	 *
	 * @param	$className	Name of the class
	 * @return	void
	 */
	protected function __construct ($className) {
		// Call parent constructor
		parent::__construct($className);

		// Calculate header size
		$this->setHeaderSize(
			strlen(self::STACK_MAGIC) +
			strlen(self::SEPARATOR_HEADER_DATA) +
			self::LENGTH_COUNT +
			strlen(self::SEPARATOR_HEADER_DATA) +
			self::LENGTH_POSITION +
			strlen(self::SEPARATOR_HEADER_ENTRIES)
		);

		// Init counters and gaps array
		$this->initCountersGapsArray();
	}

	/**
	 * Reads the file header
	 *
	 * @return	void
	 * @todo	To hard assertions here, better rewrite them to exceptions
	 */
	protected function readFileHeader () {
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] CALLED!', __METHOD__, __LINE__));

		// First rewind to beginning as the header sits at the beginning ...
		$this->getIteratorInstance()->rewind();

		// Then read it (see constructor for calculation)
		$data = $this->getIteratorInstance()->read($this->getHeaderSize());
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Read %d bytes (%d wanted).', __METHOD__, __LINE__, strlen($data), $this->getHeaderSize()));

		// Have all requested bytes been read?
		assert(strlen($data) == $this->getHeaderSize());
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Passed assert().', __METHOD__, __LINE__));

		// Last character must be the separator
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] data(-1)=%s', __METHOD__, __LINE__, dechex(ord(substr($data, -1, 1)))));
		assert(substr($data, -1, 1) == chr(self::SEPARATOR_HEADER_ENTRIES));
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Passed assert().', __METHOD__, __LINE__));

		// Okay, then remove it
		$data = substr($data, 0, -1);

		// And update seek position
		$this->updateSeekPosition();

		/*
		 * Now split it:
		 *
		 * 0 => magic
		 * 1 => total entries
		 * 2 => current seek position
		 */
		$header = explode(chr(self::SEPARATOR_HEADER_DATA), $data);

		// Set header here
		$this->setHeader($header);

		// Check if the array has only 3 elements
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] header(%d)=%s', __METHOD__, __LINE__, count($header), print_r($header, TRUE)));
		assert(count($header) == 3);
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Passed assert().', __METHOD__, __LINE__));

		// Check magic
		assert($header[0] == self::STACK_MAGIC);
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Passed assert().', __METHOD__, __LINE__));

		// Check length of count and seek position
		assert(strlen($header[1]) == self::LENGTH_COUNT);
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Passed assert().', __METHOD__, __LINE__));
		assert(strlen($header[2]) == self::LENGTH_POSITION);
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Passed assert().', __METHOD__, __LINE__));

		// Decode count and seek position
		$header[1] = hex2bin($header[1]);
		$header[2] = hex2bin($header[2]);

		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] EXIT!', __METHOD__, __LINE__));
	}

	/**
	 * Flushes the file header
	 *
	 * @return	void
	 */
	protected function flushFileHeader () {
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] CALLED!', __METHOD__, __LINE__));

		// Put all informations together
		$header = sprintf('%s%s%s%s%s%s',
			// Magic
			self::STACK_MAGIC,

			// Separator magic<->count
			chr(self::SEPARATOR_HEADER_DATA),

			// Total entries (will be zero) and pad it to 20 chars
			str_pad($this->dec2hex($this->getCounter()), self::LENGTH_COUNT, '0', STR_PAD_LEFT),

			// Separator count<->seek position
			chr(self::SEPARATOR_HEADER_DATA),

			// Position (will be zero)
			str_pad($this->dec2hex($this->getSeekPosition(), 2), self::LENGTH_POSITION, '0', STR_PAD_LEFT),

			// Separator position<->entries
			chr(self::SEPARATOR_HEADER_ENTRIES)
		);

		// Write it to disk (header is always at seek position 0)
		$this->writeData(0, $header, FALSE);

		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] EXIT!', __METHOD__, __LINE__));
	}

	/**
	 * Initializes this file-based stack.
	 *
	 * @param	$fileName	File name of this stack
	 * @param	$type		Type of this stack (e.g. url_source for URL sources)
	 * @return	void
	 * @todo	Currently the stack file is not cached, please implement a memory-handling class and if enough RAM is found, cache the whole stack file.
	 */
	protected function initFileStack ($fileName, $type) {
		// Get a stack file instance
		$fileInstance = ObjectFactory::createObjectByConfiguredName('stack_file_class', array($fileName));

		// Get iterator instance
		$iteratorInstance = ObjectFactory::createObjectByConfiguredName('file_io_iterator_class', array($fileInstance, $this));

		// Is the instance implementing the right interface?
		assert($iteratorInstance instanceof SeekableWritableFileIterator);

		// Set iterator here
		$this->setIteratorInstance($iteratorInstance);

		// Is the file's header initialized?
		if (!$this->isFileHeaderInitialized()) {
			// No, then create it (which may pre-allocate the stack)
			$this->createFileHeader();

			// And pre-allocate a bit
			$this->preAllocateFile('file_stack');
		} // END - if

		// Load the file header
		$this->readFileHeader();

		// Count all entries in file
		$this->getIteratorInstance()->analyzeFile();

		/*
		 * Get stack index instance. This can be used for faster
		 * "defragmentation" and startup.
		 */
		$indexInstance = FileStackIndexFactory::createFileStackIndexInstance($fileName, $type);

		// And set it here
		$this->setIndexInstance($indexInstance);
	}

	/**
	 * Adds a value to given stack
	 *
	 * @param	$stackerName	Name of the stack
	 * @param	$value			Value to add to this stacker
	 * @return	void
	 * @throws	FullStackerException	If the stack is full
	 */
	protected function addValue ($stackerName, $value) {
		// Do some tests
		if ($this->isStackFull($stackerName)) {
			// Stacker is full
			throw new FullStackerException(array($this, $stackerName, $value), self::EXCEPTION_STACKER_IS_FULL);
		} // END - if

		// Now add the value to the stack
		$this->partialStub('stackerName=' . $stackerName . ',value[]=' . gettype($value));
	}

	/**
	 * Get last value from named stacker
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$value			Value of last added value
	 * @throws	EmptyStackerException	If the stack is empty
	 */
	protected function getLastValue ($stackerName) {
		// Is the stack not yet initialized or full?
		if ($this->isStackEmpty($stackerName)) {
			// Throw an exception
			throw new EmptyStackerException(array($this, $stackerName), self::EXCEPTION_STACKER_IS_EMPTY);
		} // END - if

		// Now get the last value
		$this->partialStub('stackerName=' . $stackerName);
		$value = NULL;

		// Return it
		return $value;
	}

	/**
	 * Get first value from named stacker
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$value			Value of last added value
	 * @throws	EmptyStackerException	If the stack is empty
	 */
	protected function getFirstValue ($stackerName) {
		// Is the stack not yet initialized or full?
		if ($this->isStackEmpty($stackerName)) {
			// Throw an exception
			throw new EmptyStackerException(array($this, $stackerName), self::EXCEPTION_STACKER_IS_EMPTY);
		} // END - if

		// Now get the first value
		$this->partialStub('stackerName=' . $stackerName);
		$value = NULL;

		// Return it
		return $value;
	}

	/**
	 * "Pops" last entry from stack
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$value			Value "poped" from array
	 * @throws	EmptyStackerException	If the stack is empty
	 */
	protected function popLast ($stackerName) {
		// Is the stack not yet initialized or full?
		if ($this->isStackEmpty($stackerName)) {
			// Throw an exception
			throw new EmptyStackerException(array($this, $stackerName), self::EXCEPTION_STACKER_IS_EMPTY);
		} // END - if

		// Now, remove the last entry, we don't care about the return value here, see elseif() block above
		$this->partialStub('stackerName=' . $stackerName);
		return NULL;
	}

	/**
	 * "Pops" first entry from stack
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$value			Value "shifted" from array
	 * @throws	EmptyStackerException	If the named stacker is empty
	 */
	protected function popFirst ($stackerName) {
		// Is the stack not yet initialized or full?
		if ($this->isStackEmpty($stackerName)) {
			// Throw an exception
			throw new EmptyStackerException(array($this, $stackerName), self::EXCEPTION_STACKER_IS_EMPTY);
		} // END - if

		// Now, remove the last entry, we don't care about the return value here, see elseif() block above
		$this->partialStub('stackerName=' . $stackerName);
		return NULL;
	}

	/**
	 * Checks whether the given stack is full
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$isFull			Whether the stack is full
	 */
	protected function isStackFull ($stackerName) {
		// File-based stacks will only run full if the disk space is low.
		// @TODO Please implement this, returning FALSE
		$isFull = FALSE;

		// Return result
		return $isFull;
	}

	/**
	 * Checks whether the given stack is empty
	 *
	 * @param	$stackerName		Name of the stack
	 * @return	$isEmpty			Whether the stack is empty
	 * @throws	NoStackerException	If given stack is missing
	 */
	public function isStackEmpty ($stackerName) {
		// So, is the stack empty?
		$isEmpty = (($this->getStackCount($stackerName)) == 0);

		// Return result
		return $isEmpty;
	}

	/**
	 * Initializes given stacker
	 *
	 * @param	$stackerName	Name of the stack
	 * @param	$forceReInit	Force re-initialization
	 * @return	void
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function initStack ($stackerName, $forceReInit = FALSE) {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Initializes all stacks
	 *
	 * @return	void
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function initStacks (array $stacks, $forceReInit = FALSE) {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Checks whether the given stack is initialized (set in array $stackers)
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$isInitialized	Whether the stack is initialized
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function isStackInitialized ($stackerName) {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Getter for size of given stack (array count)
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$count			Size of stack (array count)
	 */
	public function getStackCount ($stackerName) {
		// Now, simply return the found count value, this must be up-to-date then!
		return $this->getCounter();
	}

	/**
	 * Calculates minimum length for one entry/block
	 *
	 * @return	$length		Minimum length for one entry/block
	 */
	public function caluclateMinimumBlockLength () {
		// Calulcate it
		$length = self::getHashLength() + strlen(chr(self::SEPARATOR_HASH_NAME)) + self::LENGTH_NAME + 1 + strlen(self::getBlockSeparator());

		// Return it
		return $length;
	}
}

// [EOF]
?>
